import argparse
import json
import os

import optuna

from mu_alpha_zero.AlphaZero.utils import az_optuna_parameter_search

os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"


def main():
    parser_ = argparse.ArgumentParser(
        description="Start optuna optimization jobs"
    )
    # parser_.add_argument("n", help="Number of processes to start.")
    parser_.add_argument("t", help="Number of trials to run.")
    parser_.add_argument("-s", "--storage", help="The storage string to use.")
    parser_.add_argument("-n", "--study_name", help="The name of the optuna study to use.")
    parser_.add_argument("-i", "--init_net_path", help="The path to the initial network.")

    args = parser_.parse_args()
    # n = int(args.n)
    t = int(args.t)
    storage = args.storage
    if storage is None:
        storage = "mysql://root:584792@localhost/alpha_zero"
    study_name = args.study_name
    if study_name is None:
        study_name = "alpha_zero"

    init_net_path = args.init_net_path

    # print(f"Starting optuna optimization. Using parameters: \n"
    #       f"Number of parallel processes: {n}\n"
    #       f"Number of trials: {t}\n"
    #       f"Storage: {storage}\n"
    #       f"Study name: {study_name}\n"
    #       f"Starting...")
    az_optuna_parameter_search(n_trials=t, init_net_path=init_net_path, storage=storage, study_name=study_name)

    study = optuna.load_study(study_name=study_name, storage=storage)
    with open("best_params.json", "w") as file:
        json.dump(study.best_params, file)


if __name__ == "__main__":
    main()



import random
from collections import deque
from itertools import chain

import numpy as np
import pymongo
import torch as th
from diskcache import Deque
from torch.utils.data import Dataset, DataLoader

from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.utils import find_project_root
from mu_alpha_zero.MuZero.utils import scale_action
from mu_alpha_zero.MuZero.pickler import DataPickler


class MemDataset(Dataset):
    def __init__(self, mem_buffer):
        self.mem_buffer = list(mem_buffer)

    def __len__(self):
        return len(self.mem_buffer)

    def __getitem__(self, idx):
        return self.mem_buffer[idx]


class MemBuffer(GeneralMemoryBuffer):
    def __init__(self, max_size, disk: bool = False, dir_path: str = None):
        self.max_size = max_size
        self.disk = disk
        self.dir_path = dir_path
        self.buffer = self.init_buffer(dir_path)
        self.last_buffer_size = 0
        self.priorities = None
        self.is_disk = disk

    def add(self, experience):
        if not isinstance(experience, tuple):
            raise ValueError("Experience must be a tuple")
        self.buffer.append(experience)

    def init_buffer(self, dir_path: str or None):
        if self.disk:
            if dir_path is None:
                dir_path = f"{find_project_root()}/Pickles/Data"
            return Deque(maxlen=self.max_size, directory=dir_path)
        else:
            return deque(maxlen=self.max_size)

    def add_list(self, experience_list):
        self.buffer.extend(experience_list)

    def sample(self, batch_size):
        return random.sample(self.buffer, batch_size)

    def shuffle(self):
        # if self.disk:
        #     raise NotImplementedError("Shuffle not implemented for disk buffer")
        random.shuffle(self.buffer)  # in-place shuffle

    def batch(self, batch_size):
        batched_buffer = []
        buffer_len = len(self.buffer)
        for idx in range(0, buffer_len, batch_size):
            batched_buffer.append(list(self.buffer)[idx:min(idx + batch_size, buffer_len)])

        return batched_buffer

    def __call__(self, batch_size) -> list:
        return self.batch(batch_size)

    def __len__(self):
        return len(self.buffer)

    def batch_with_priorities(self, epochs, batch_size, K, alpha=1):
        for _ in range(epochs):
            if self.last_buffer_size < len(self.buffer):
                priorities = self.calculate_priorities(batch_size, alpha, K)
                self.priorities = priorities
                self.last_buffer_size = len(self.buffer)
            random_indexes = np.random.choice(np.arange(len(self.buffer) - K),
                                              size=min(len(priorities) // K, max(batch_size // K, 1)),
                                              replace=False, p=self.priorities).tolist()
            batch = [[self.buffer[b] for b in range(i, i + K)] for i in random_indexes]
            pris = [self.priorities[i:i + K] for i in random_indexes]
            yield list(chain.from_iterable(batch)), th.tensor(list(chain.from_iterable(pris)), dtype=th.float32)

    def calculate_priorities(self, batch_size, alpha, K):
        ps = [abs(self.buffer[i][1] - self.buffer[i][2][2]) for i in
              range(len(self.buffer))][:-K]
        ps = np.array(ps)
        ps = ps ** alpha
        ps = ps / ps.sum()
        return ps

    def make_fresh_instance(self):
        return MemBuffer(self.max_size, self.disk, self.dir_path)

    def to_dataloader(self, batch_size):
        return DataLoader(MemDataset(self.buffer), batch_size=batch_size, shuffle=True, collate_fn=lambda x: x)

    def save(self):
        if not self.disk:
            buffer = list(self.buffer)
            th.save(buffer, f"{find_project_root()}/Pickles/memory_buffer.pt")
            return f"{find_project_root()}/Pickles/memory_buffer.pt"
        else:
            return self.buffer.directory


class MuZeroFrameBuffer:
    def __init__(self, frame_buffer_size, noop_action: int, action_space_size: int):
        self.max_size = frame_buffer_size
        self.noop_action = noop_action
        self.action_space_size = action_space_size
        self.buffer = deque(maxlen=frame_buffer_size)

    def add_frame(self, frame, action):
        self.buffer.append((frame, action))

    def concat_frames(self):
        frames_with_actions = [th.cat((th.tensor(frame, dtype=th.float32),
                                       th.full((frame.shape[0], frame.shape[1], 1), action, dtype=th.float32)), dim=2)
                               for frame, action in self.buffer]
        # return th.tensor(np.array(frames_with_actions), dtype=th.float32)
        return th.cat(frames_with_actions, dim=2)

    def init_buffer(self, init_state):
        for _ in range(self.max_size):
            self.add_frame(init_state, scale_action(self.noop_action, self.action_space_size))

    def __len__(self):
        return len(self.buffer)


class MongoDBMemBuffer(GeneralMemoryBuffer):
    def __init__(self):
        self.db = pymongo.MongoClient("localhost", 27017).muzero
        self.calculated_buffer_size = 0
        self.is_disk = False

    def add(self, experience):
        if not isinstance(experience, dict):
            raise ValueError("Experience must be a dict")
        self.db.game_data.insert(experience)

    def add_list(self, experience_list):
        self.db.game_data.insert_many(experience_list)

    def batch(self, batch_size):
        random_idx = random.randint(0, self.db.game_data.count_documents({}) - batch_size)
        return list(self.db.game_data.find({}).skip(random_idx).limit(batch_size))

    def calculate_priorities(self, batch_size, alpha, K):
        self.calculated_buffer_size = self.db.game_data.count_documents({})
        fields = self.db.game_data.find({}, {"_id": 0, "pred_reward": 1, "t_reward": 1})
        ps = [abs(x["pred_reward"] - x["t_reward"]) ** alpha for x in fields]
        # add ps to db
        document_ids = self.db.game_data.find({}, {"_id": 1})
        for doc_id, p in zip(document_ids, ps):
            self.db.game_data.update_one(doc_id, {"$set": {"priority": p}})

    def update_priorities_if_needed(self, alpha, K):
        if self.calculated_buffer_size < self.db.game_data.count_documents({}):
            self.calculate_priorities(self.calculated_buffer_size, alpha, K)

    def batch_with_priorities(self, epochs, batch_size, K, alpha=1):
        for _ in range(epochs):
            self.update_priorities_if_needed(alpha, K)
            test_p = list(self.db.game_data.find({}, {"priority": 1, "_id": 0}).limit(3))
            # test_p = list(test_p)[0]["priority"]
            priorities = [x["priority"] for x in self.db.game_data.find({}, {"priority": 1, "_id": 0})]
            sum_p = sum(priorities)
            priorities = [p / sum_p for p in priorities]
            indexes = np.random.choice(np.arange(self.db.game_data.count_documents({})),
                                       size=min(self.calculated_buffer_size, batch_size // K), replace=False,
                                       p=priorities).tolist()
            items = [list(self.db.game_data.find({}).skip(x).limit(K)) for x in indexes]
            items = list(chain.from_iterable(items))
            items = tuple(
                [(x["probabilities"], x["vs"], (x["t_reward"], x["game_move"], x["pred_reward"]), x["game_state"]) for x
                 in
                 items])
            yield items, th.tensor(priorities, dtype=th.float32)

    def get_last_greatest_id(self):
        return self.db.game_data.find_one(sort=[("_id", pymongo.DESCENDING)])["_id"]

    def __len__(self):
        return self.db.game_data.count_documents({})

    def drop_game_data(self):
        self.db.game_data.drop()

    def make_fresh_instance(self):
        return MongoDBMemBuffer()


class PickleMemBuffer(GeneralMemoryBuffer):

    def __init__(self, pickle_dir: str):
        self.pickle_dir = pickle_dir
        self.is_disk = True
        self.pickler = DataPickler(pickle_dir)

    def add(self, experience):
        raise NotImplementedError("Single experience addition shouldn't be performed in PickleMemBuffer, please add "
                                  "entire bach with add_list method")

    def add_list(self, experience_list):
        self.pickler.pickle_buffer(experience_list)
        # self.pickler.push_to_consumer(experience_list)

    def batch(self, batch_size):
        raise NotImplementedError("Batch method not implemented for PickleMemBuffer, please use batch_with_priorities")

    def calculate_priorities(self, batch_size, alpha, K):
        vs = self.pickler.load_index(1)
        rews = self.pickler.load_index(2)
        ps = [(abs(vs[i] - rews[i][2]) ** alpha, i) for i in range(len(vs))][:-K]
        sum_p = sum([p[0] for p in ps])
        ps = [(p[0] / sum_p, p[1]) for p in ps]
        return {p[1]: p[0] for p in ps}

    def batch_with_priorities(self, epochs, batch_size, K, alpha=1):
        priorities = self.calculate_priorities(batch_size, alpha, K)
        ps_probs = np.array(list(priorities.values()))
        for _ in range(epochs):
            random_indexes = np.random.choice(np.arange(len(priorities)),
                                              size=min(len(priorities) // K, max(batch_size // K, 1)),
                                              replace=False, p=list(ps_probs)).tolist()
            batch = self.pickler.load_all(batch_size, random_indexes, K)
            pris = [list(priorities.values())[i:i + K] for i in random_indexes]
            tmp = list(chain.from_iterable(pris))
            yield batch, th.tensor(tmp, dtype=th.float32)

    def save(self):
        print("Load using datapickler.load_all(float(inf),...)")
        return self.pickle_dir

    def make_fresh_instance(self):
        return PickleMemBuffer(self.pickle_dir)

    def __len__(self):
        raise NotImplementedError("Length not implemented for PickleMemBuffer.")



import atexit
import glob
import os

import numpy as np
import torch as th
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.functional import mse_loss

from mu_alpha_zero.General.network import GeneralAlphZeroNetwork
from mu_alpha_zero.mem_buffer import MemBuffer
from mu_alpha_zero.config import AlphaZeroConfig


class AlphaZeroNet(nn.Module, GeneralAlphZeroNetwork):
    def __init__(self, in_channels: int, num_channels: int, dropout: float, action_size: int, linear_input_size: int):
        super(AlphaZeroNet, self).__init__()
        self.in_channels = in_channels
        self.num_channels = num_channels
        self.dropout_p = dropout
        self.action_size = action_size
        self.linear_input_size = linear_input_size

        # Convolutional layers
        self.conv1 = nn.Conv2d(in_channels, num_channels, 3, padding=1)
        self.bn1 = nn.BatchNorm2d(num_channels)
        self.conv2 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.bn2 = nn.BatchNorm2d(num_channels)
        self.conv3 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.bn3 = nn.BatchNorm2d(num_channels)
        self.conv4 = nn.Conv2d(num_channels, num_channels, 3)
        self.bn4 = nn.BatchNorm2d(num_channels)

        # Fully connected layers
        # 4608 (5x5) or 512 (3x3) or 32768 (10x10) or 18432 (8x8) # or 8192 for atari (6x6)
        self.fc1 = nn.Linear(self.linear_input_size, 1024)
        self.fc1_bn = nn.BatchNorm1d(1024)
        self.fc2 = nn.Linear(1024, 512)
        self.fc2_bn = nn.BatchNorm1d(512)

        self.dropout = nn.Dropout(dropout)

        # Output layers
        self.pi = nn.Linear(512, action_size)  # probability head
        self.v = nn.Linear(512, 1)  # value head
        atexit.register(self.clear_traces)

    def forward(self, x, muzero=True):
        if not muzero:
            x = x.unsqueeze(1)
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))
        x = F.relu(self.bn4(self.conv4(x)))

        x = x.reshape(x.size(0), -1)  # Flatten

        x = F.relu(self.fc1_bn(self.fc1(x)))
        x = self.dropout(x)

        x = F.relu(self.fc2_bn(self.fc2(x)))
        x = self.dropout(x)

        pi = F.log_softmax(self.pi(x), dim=1)
        v = F.tanh(self.v(x))

        return pi, v

    @th.no_grad()
    def predict(self, x, muzero=True):
        pi, v = self.forward(x, muzero=muzero)
        pi = th.exp(pi)
        return pi.detach().cpu().numpy(), v.detach().cpu().numpy()

    def to_traced_script(self, board_size: int = 10):
        return th.jit.trace(self, th.rand(1, 256, board_size, board_size).cuda())

    def trace(self, board_size: int) -> str:
        traced = self.to_traced_script(board_size=board_size)
        path = "traced.pt"
        traced.save(path)
        return path

    def clear_traces(self) -> None:
        from mu_alpha_zero.General.utils import find_project_root
        for trace_file in glob.glob(f"{find_project_root()}/Checkpoints/Traces/*.pt"):
            os.remove(trace_file)

    def make_fresh_instance(self):
        return AlphaZeroNet(self.in_channels, self.num_channels, self.dropout_p, self.action_size,
                            self.linear_input_size)

    @staticmethod
    def make_from_config(config: AlphaZeroConfig):
        return AlphaZeroNet(config.num_net_in_channels, config.num_net_channels, config.net_dropout,
                            config.net_action_size, config.az_net_linear_input_size)

    def train_net(self, memory_buffer: MemBuffer, alpha_zero_config: AlphaZeroConfig):
        from mu_alpha_zero.AlphaZero.utils import mask_invalid_actions_batch
        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        losses = []
        optimizer = th.optim.Adam(self.parameters(), lr=alpha_zero_config.lr)
        memory_buffer.shuffle()
        for epoch in range(alpha_zero_config.epochs):
            for experience_batch in memory_buffer(alpha_zero_config.batch_size):
                if len(experience_batch) <= 1:
                    continue
                states, pi, v = zip(*experience_batch)
                states = th.tensor(np.array(states), dtype=th.float32, device=device)
                pi = th.tensor(np.array(pi), dtype=th.float32, device=device)
                v = th.tensor(v, dtype=th.float32, device=device).unsqueeze(1)
                pi_pred, v_pred = self.forward(states, muzero=False)
                masks = mask_invalid_actions_batch(states)
                loss = mse_loss(v_pred, v) + self.pi_loss(pi_pred, pi, masks, device)
                losses.append(loss.item())
                # self.summary_writer.add_scalar("Loss", loss.item(), i * epochs + epoch)

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

        return sum(losses) / len(losses), losses

    def pi_loss(self, y_hat, y, masks, device: th.device):
        masks = masks.reshape(y_hat.shape).to(device)
        masked_y_hat = masks * y_hat
        return -th.sum(y * masked_y_hat) / y.size()[0]

    def to_shared_memory(self):
        for param in self.parameters():
            param.share_memory_()


class TicTacToeNetNoNorm(nn.Module):
    def __init__(self, in_channels, num_channels, dropout, action_size):
        super(TicTacToeNetNoNorm, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, num_channels, 3, padding=1)
        self.conv2 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.conv3 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.conv4 = nn.Conv2d(num_channels, num_channels, 3)

        self.fc1 = nn.Linear(4608, 1024)
        self.fc2 = nn.Linear(1024, 512)

        self.dropout = nn.Dropout(dropout)

        self.pi = nn.Linear(512, action_size)  # probability head
        self.v = nn.Linear(512, 1)  # value head

    def forward(self, x):
        x = x.unsqueeze(1)
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        x = F.relu(self.conv3(x))
        x = F.relu(self.conv4(x))

        x = x.view(x.size(0), -1)

        x = F.relu(self.fc1(x))
        x = self.dropout(x)

        x = F.relu(self.fc2(x))
        x = self.dropout(x)

        pi = F.softmax(self.pi(x), dim=1)
        v = F.tanh(self.v(x))
        # print(th.any(th.isnan(pi)))
        # print(th.any(th.isnan(v)))

        return pi, v

    def predict(self, x):
        pi, v = self.forward(x)
        return pi.detach().cpu().numpy(), v.detach().cpu().numpy()



import numbers
import os
import shutil
import subprocess

import numpy as np
from py4j.java_gateway import JavaGateway
from mu_alpha_zero.General.utils import get_python_home, find_project_root, get_site_packages_path
from mu_alpha_zero.MuZero.JavaGateway.java_networks import JavaNetworks
from mu_alpha_zero.MuZero.Network.networks import MuZeroNet
import atexit


class JavaManager:
    def __init__(self, network: MuZeroNet, args: dict):
        # jpype.addClassPath(f"{find_project_root()}/MuZero/JavaGateway/JSeflplayLinux-1.0-SNAPSHOT.jar")
        # jpype.addClassPath(f"/home/skyr/IdeaProjects/JSelfPlay/target/JSeflplayLinux-1.0-SNAPSHOT.jar")
        os.environ["PYTHONHOME"] = get_python_home()
        # jpype.startJVM(jpype.getDefaultJVMPath(),f"-Djava.library.path={get_site_packages_path()}/jep/")
        self.java_nets = JavaNetworks()
        self.net = network
        self.env_id = args["env_id"]
        self.args = self.prepare_args(args)
        self.java_process = self.spawn_java_p()
        self.java_gateway = JavaGateway()
        atexit.register(self.kill_java_process)

    def prepare_args(self, args: dict):
        return {k: v for k, v in args.items() if
                isinstance(v, numbers.Number) and not isinstance(v, bool) and v != float("inf") and v != float("-inf")}

    def spawn_java_p(self):
        return subprocess.Popen(
            ["java", "-jar", "/home/skyr/IdeaProjects/JSelfPlay/target/JSeflplayLinux-1.0-SNAPSHOT.jar",
             f"-Djava.library.path={get_site_packages_path()}/jep/"])

    def run_parallel_java_self_play(self, n_jobs: int, n_games: int):
        results = []
        path = self.java_nets.save(self.net)
        os.makedirs(f"{find_project_root()}/Arrays/", exist_ok=True)
        res = self.java_gateway.entry_point.runParallelSelfPlay(n_jobs, n_games, path, self.dict_to_java_map(self.args),
                                                                self.env_id)
        arr_paths = set([x.getValue3() for x in res])
        arrs = []
        for path in arr_paths:
            arrs.append(np.load(path))
        arr = np.concatenate(arrs, axis=0)
        for i,quarted in enumerate(res):
            hmap = quarted.getValue0()
            pd = {int(entry.getKey()): float(entry.getValue()) for entry in hmap.entrySet()}
            v = float(quarted.getValue1())
            rmpred_r = (float(quarted.getValue2().getValue0()), float(quarted.getValue2().getValue1()),
                        float(quarted.getValue2().getValue2()))
            results.append((pd, v, rmpred_r, arr[i]))
        shutil.rmtree(f"{find_project_root()}/Arrays/")
        return results

    def dict_to_java_map(self, dict: dict):
        java_map = self.java_gateway.jvm.java.util.HashMap()
        for k, v in dict.items():
            java_map.put(k, str(v))
        return java_map

    def kill_java_process(self):
        self.java_process.kill()
        self.java_process.wait()
        print("Java process ended.")
        return



import math

from mu_alpha_zero.MuZero.Network.networks import MuZeroNet
from mu_alpha_zero.General.utils import find_project_root


class JavaNetworks:
    def save(self, networks_wrapper: MuZeroNet) -> str:
        path = f"{find_project_root()}/mz_net.pth"
        networks_wrapper.to_pickle(path)
        return path



from abc import ABC, abstractmethod

import numpy as np


class AlphaZeroGame(ABC):
    """
    Make your custom game extend this class and implement all the methods.
    """

    def game_result(self, player, board) -> float | None:
        """
        Returns the result of the game from the perspective of the supplied player.
        :param player: The player to check for (1 or -1).
        :param board: The board to check on. If None, the current board is used.
        :return: The game result. None when the game is not over yet.
        """
        if self.check_win(player, board=board):
            return 1.0
        if self.check_win(-player, board=board):
            return -1.0
        if self.is_board_full(board=board):
            return 1e-4
        return None

    @staticmethod
    def select_move(action_probs: dict):
        moves, probs = zip(*action_probs.items())
        return np.random.choice(moves, p=probs)

    @abstractmethod
    def make_fresh_instance(self):
        """
        Returns a fresh instance of the game (not a copy).
        """
        pass

    @abstractmethod
    def check_win(self, player: int, board: np.ndarray) -> bool:
        """
        Checks if the current player 'player' won on the 'board'. If so returns true, else false.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_canonical_form(self, board: np.ndarray, player: int) -> np.ndarray:
        """
        Returns the canonical form of the board.
        In this case, that is a board where the current player is always player 1.
        You probably want: return board * player
        """
        pass

    @abstractmethod
    def get_next_state(self, board: np.ndarray, action: int or tuple, player: int) -> np.ndarray:
        """
        This method plays move given by 'board_index' as 'player' on the given 'board'
        and returns the updated board.
        """
        pass

    @abstractmethod
    def is_board_full(self, board: np.ndarray) -> bool:
        """
        Checks if the given board is completely full. Returns true if so, false other otherwise,
        """
        pass

    @abstractmethod
    def get_valid_moves(self, board: np.ndarray, player: int or None) -> np.ndarray:
        """
        Returns a list of valid moves for the given player on the given board.
        If valid moves are independent of player, use None for player.
        """
        pass

    @abstractmethod
    def get_board(self) -> np.ndarray:
        """
        Returns a copy of the current board.
        """
        pass

    @abstractmethod
    def reset(self):
        """
        Resets the game and returns the initial board.
        """
        pass

    def eval_board(self, board: np.ndarray, player: int):
        """
        Optional. Evaluates the board from the perspective of the given player.
        This is used for minimax, override this method if you want to use it.
        """
        pass

    @abstractmethod
    def render(self):
        """
        Render the GUI of the game.
        """



import sys
from typing import Type

import numpy as np
import torch as th

from mu_alpha_zero.AlphaZero.Arena.arena import Arena
from mu_alpha_zero.AlphaZero.Arena.players import NetPlayer
from mu_alpha_zero.AlphaZero.MCTS.az_search_tree import McSearchTree
from mu_alpha_zero.AlphaZero.Network.trainer import Trainer
from mu_alpha_zero.General.az_game import AlphaZeroGame
from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.network import GeneralNetwork
from mu_alpha_zero.General.utils import net_not_none, find_project_root
from mu_alpha_zero.config import AlphaZeroConfig


class AlphaZero:
    def __init__(self, game_instance: AlphaZeroGame):
        self.trainer = None
        self.net = None
        self.game = game_instance
        self.device = th.device("cuda" if th.cuda.is_available() else "cpu")
        self.alpha_zero_config: AlphaZeroConfig = None
        self.tree: McSearchTree = None

    def create_new(self, alpha_zero_config: AlphaZeroConfig, network_class: Type[GeneralNetwork],
                   memory: GeneralMemoryBuffer,
                   headless: bool = True,
                   checkpointer_verbose: bool = False):
        network = network_class.make_from_config(alpha_zero_config).to(self.device)
        tree = McSearchTree(self.game.make_fresh_instance(), alpha_zero_config)
        self.tree = tree
        net_player = NetPlayer(self.game.make_fresh_instance(), **{"network": network, "monte_carlo_tree_search": tree})
        self.alpha_zero_config = alpha_zero_config
        self.trainer = Trainer.create(alpha_zero_config, self.game, network, tree, net_player, headless=headless,
                                      checkpointer_verbose=checkpointer_verbose, memory_override=memory)
        self.net = self.trainer.get_network()

    def load_checkpoint(self, network_class: Type[GeneralNetwork], path: str, checkpoint_dir: str,
                        headless: bool = True,
                        checkpointer_verbose: bool = False):
        self.trainer = Trainer.from_checkpoint(network_class, McSearchTree, NetPlayer, path, checkpoint_dir, self.game,
                                               headless=headless,
                                               checkpointer_verbose=checkpointer_verbose)
        self.net = self.trainer.get_network()
        self.tree = self.trainer.get_tree()
        self.args = self.trainer.get_args()

    def train(self):
        net_not_none(self.net)
        self.trainer.train()

    def predict(self, x: np.ndarray, tau: float = 0) -> int:
        net_not_none(self.net)
        assert x.shape == (self.args["board_size"], self.args["board_size"], self.args["num_net_in_channels"]), \
            "Input shape is not correct. Expected (board_size, board_size, num_net_in_channels)." \
            "Got: " + str(x.shape)
        pi, _ = self.tree.search(self.net, x, 1, self.device, tau=tau)
        return self.game.select_move(pi)

    def play(self, p1_name: str, p2_name: str, num_games: int, alpha_zero_config: AlphaZeroConfig, starts: int = 1,
             switch_players: bool = True):
        net_not_none(self.net)
        self.net.to(self.device)
        self.net.eval()
        manager = self.game.make_fresh_instance()
        tree = McSearchTree(manager, alpha_zero_config)
        kwargs = {"network": self.net, "monte_carlo_tree_search": tree, "evaluate_fn": manager.eval_board,
                  "depth": alpha_zero_config.minimax_depth, "player": -1}
        path_prefix = find_project_root().replace("\\", "/").split("/")[-1]
        p1 = sys.modules[f"{path_prefix}.AlphaZero.Arena.players"].__dict__[p1_name](manager, **kwargs)
        p2 = sys.modules[f"{path_prefix}.AlphaZero.Arena.players"].__dict__[p2_name](manager, **kwargs)
        arena_manager = self.game.make_fresh_instance()
        arena_manager.set_headless(False)
        arena = Arena(arena_manager, alpha_zero_config, self.device)
        p1_w, p2_w, ds = arena.pit(p1, p2, num_games, alpha_zero_config.num_simulations, one_player=not switch_players,
                                   start_player=starts, add_to_kwargs=kwargs)
        print(f"Results: Player 1 wins: {p1_w}, Player 2 wins: {p2_w}, Draws: {ds}")




import os

import ray

from AlphaZero.MCTS.az_search_tree import McSearchTree

os.environ['OMP_NUM_THREADS'] = '5'
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
from AlphaZero.Network.trainer import Trainer

from AlphaZero.constants import SAMPLE_AZ_ARGS as args_
from AlphaZero.constants import TRAINED_AZ_NET_ARGS as _args
from AlphaZero.utils import az_optuna_parameter_search, DotDict, make_net_from_checkpoint, build_net_from_config
import argparse
from AlphaZero.Arena.players import HumanPlayer, NetPlayer, JavaMinimaxPlayer
from Game.tictactoe_game import TicTacToeGameManager as GameManager
from AlphaZero.Arena.arena import Arena
from MuZero.utils import mz_optuna_parameter_search as mu_optuna_parameter_search
import torch as th


def main_alpha_zero_find_hyperparameters():
    parser_ = argparse.ArgumentParser()
    parser_.add_argument("-s", "--storage", help="The storage string to use.")
    parser_.add_argument("-n", "--study_name", help="The name of the optuna study to use.")
    parser_.add_argument("-t", "--n_trials", help="The number of trials to run.")
    parser_.add_argument("-i", "--init_net_path", help="The path to the initial network.")

    args_ = parser_.parse_args()
    storage = args_.storage
    study_name = args_.study_name
    n_trials = int(args_.n_trials)
    init_net_path = args_.init_net_path

    az_optuna_parameter_search(n_trials=n_trials, init_net_path=init_net_path,
                               storage=storage, study_name=study_name)


def main_mu_zero_find_hyperparameters():
    parser_ = argparse.ArgumentParser()
    parser_.add_argument("-s", "--storage", help="The storage string to use.")
    parser_.add_argument("-n", "--study_name", help="The name of the optuna study to use.")
    parser_.add_argument("-t", "--n_trials", help="The number of trials to run.")
    parser_.add_argument("-i", "--init_net_path", help="The path to the initial network.")

    args_ = parser_.parse_args()
    storage = args_.storage
    study_name = args_.study_name
    n_trials = int(args_.n_trials)
    init_net_path = args_.init_net_path

    mu_optuna_parameter_search(n_trials=n_trials, init_net_path=init_net_path,
                               storage=storage, study_name=study_name)


def main_alpha_zero():
    ray.init(num_gpus=1)
    for file_name in os.listdir("Logs/AlphaZero"):
        os.remove(f"Logs/AlphaZero/{file_name}")

    args = DotDict(args_)

    # !!! Use found args !!!

    args.num_simulations = 1317
    args.self_play_games = 300
    args.epochs = 500
    args.lr = 0.0032485504583772953
    args.tau = 1.0
    args.c = 1
    args.arena_tau = 0.04139160592420218
    # args.arena_tau = args.tau
    # args.log_epsilon = 1.4165210108199043e-08

    args.num_iters = 50
    game = GameManager(args.board_size, headless=True, num_to_win=args.num_to_win)
    # trainer = Trainer.create(args)
    tree = McSearchTree(game, args)
    network = build_net_from_config(args, th.device("cuda") if th.cuda.is_available() else th.device("cpu"))
    net_player = NetPlayer(game, **{"network": network, "monte_carlo_tree_search": tree})
    trainer = Trainer.create(args, game, network=network, net_player=net_player, search_tree=tree)
    trainer.train()
    trainer.save_latest("Checkpoints/AlphaZero/latest.pth")


def play():
    args = DotDict(args_)
    args.num_simulations = 1317
    args.self_play_games = 300
    args.epochs = 500
    args.lr = 0.0032485504583772953
    args.tau = 1.0
    args.c = 1
    args.arena_tau = 0.04139160592420218
    # args.num_to_win = 3
    # args.board_size = 5
    # args.net_action_size = args.board_size ** 2
    net = make_net_from_checkpoint(
        r"C:\Users\Skyr\PycharmProjects\AlphaZeroTicTacToe\Nets\8x8_test\improved_net_12.pth",
        args)
    net.eval()
    manager = GameManager(8, headless=False, num_to_win=5)
    search_tree = McSearchTree(manager, args)
    # p1 = TrainingNetPlayer(net, manager, args)
    p1 = NetPlayer(manager, **{"network": net, "monte_carlo_tree_search": search_tree})
    # p1 = TrainingNetPlayer(net,manager,args)
    # opp_data = th.load("/home/skyr/Downloads/temp_net_119.pth")
    # opp_net = build_net_from_args(args, th.device("cuda"))
    # # opp_net.load_state_dict(opp_data)
    # manager2 = GameManager(8, headless=False, num_to_win=5)
    # search_tree2 = McSearchTree(manager2, args)
    # p2 = RandomPlayer(manager,**{})
    p2 = HumanPlayer(manager, **{})
    # p2 = NetPlayer(opp_net,search_tree2,manager2)
    device = th.device("cuda" if th.cuda.is_available() else "cpu")
    arena = Arena(manager, args, device)
    # net.eval()
    # opp_net.eval()
    net_wins, human_wins, draws = arena.pit(p1, p2, num_games_to_play=10, num_mc_simulations=1317, one_player=False,
                                            start_player=-1, debug=True)
    print(f"Net wins: {net_wins}, Human wins: {human_wins}, Draws: {draws}")


def human_minimax_play():
    args = DotDict(_args)
    manager = GameManager(8, headless=False, num_to_win=5)
    p1 = HumanPlayer(manager.make_fresh_instance(), **{})
    new_man = manager.make_fresh_instance()
    p2 = JavaMinimaxPlayer(new_man,
                           **{"evaluate_fn": new_man.eval_board})
    device = th.device("cuda" if th.cuda.is_available() else "cpu")
    arena = Arena(manager.make_fresh_instance(), args, device)
    net_wins, human_wins, draws = arena.pit(p1, p2, num_games_to_play=10, num_mc_simulations=1317, one_player=True,
                                            start_player=1, add_to_kwargs={"depth": 5, "player": -1}, debug=True)


def ai_minimax_play():
    args = DotDict(args_)
    args.num_simulations = 1317
    args.self_play_games = 300
    args.epochs = 500
    args.lr = 0.0032485504583772953
    args.tau = 1.0
    args.c = 1
    args.arena_tau = 0.04139160592420218
    net = make_net_from_checkpoint(
        r"C:\Users\Skyr\PycharmProjects\AlphaZeroTicTacToe\Nets\3x3\improved_net_2.pth",
        args)
    net.eval()
    manager = GameManager(3, headless=False, num_to_win=3)
    search_tree = McSearchTree(manager.make_fresh_instance(), args)
    p1 = NetPlayer(manager.make_fresh_instance(), **{"network": net, "monte_carlo_tree_search": search_tree})
    new_manager = manager.make_fresh_instance()
    p2 = JavaMinimaxPlayer(new_manager,
                           **{"evaluate_fn": new_manager.eval_board})
    device = th.device("cuda" if th.cuda.is_available() else "cpu")
    arena = Arena(manager.make_fresh_instance(), args, device)
    net_wins, human_wins, draws = arena.pit(p1, p2, num_games_to_play=100, num_mc_simulations=1317, one_player=False,
                                            start_player=-1, add_to_kwargs={"depth": 5, "player": -1}, debug=False)
    print(f"Net wins: {net_wins}, Minimax wins: {human_wins}, Draws: {draws}")


if __name__ == "__main__":
    # human_minimax_play()
    # main_mu_zero_find_hyperparameters()
    # play()
    ai_minimax_play()
    # main_alpha_zero()
    # main_alpha_zero_find_hyperparameters()



import random
from typing import Type

import torch as th

from mu_alpha_zero.AlphaZero.Arena.players import Player
from mu_alpha_zero.General.arena import GeneralArena
from mu_alpha_zero.General.mz_game import MuZeroGame
from mu_alpha_zero.MuZero.utils import resize_obs
from mu_alpha_zero.config import MuZeroConfig


class MzArena(GeneralArena):
    def __init__(self, game_manager: MuZeroGame, muzero_config: MuZeroConfig, device: th.device):
        self.game_manager = game_manager
        self.muzero_config = muzero_config
        self.device = device

    def pit(self, player1: Type[Player], player2: Type[Player], num_games_to_play: int, num_mc_simulations: int,
            one_player: bool = False, start_player: int = 1):
        tau = self.muzero_config.arena_tau
        rewards = {1: [], -1: []}
        if one_player:
            num_games_per_player = num_games_to_play
        else:
            num_games_per_player = num_games_to_play // 2
        noop_num = random.randint(0, 30)
        for player in [1, -1]:
            kwargs = {"num_simulations": num_mc_simulations, "current_player": player, "device": self.device,
                      "tau": tau, "unravel": False}
            for game in range(num_games_per_player):
                self.game_manager.reset()

                state, _, _ = self.game_manager.frame_skip_step(self.game_manager.get_noop(), None,
                                                                frame_skip=noop_num)
                state = resize_obs(state, self.muzero_config.target_resolution)
                for step in range(self.muzero_config.num_steps):
                    self.game_manager.render()
                    if player == 1:
                        move = player1.choose_move(state, **kwargs)
                    else:
                        move = player2.choose_move(state, **kwargs)
                    state, reward, done = self.game_manager.frame_skip_step(move, None)
                    state = resize_obs(state, self.muzero_config.target_resolution)
                    rewards[player].append(reward)
                    if done:
                        break

        return sum(rewards[1]), sum(rewards[-1]), 0



import atexit
import datetime
import logging
import os

from pushbullet import API

from mu_alpha_zero.General.utils import find_project_root


class Logger:
    def __init__(self, logdir: str = "Logs", token: str or None = None) -> None:
        self.logdir = f"{find_project_root()}/{logdir}/ProgramLogs"
        os.makedirs(self.logdir, exist_ok=True)
        self.logger = logging.getLogger("AlphaZeroLogger")
        self.logger.setLevel(logging.DEBUG)
        self.file_handler = logging.FileHandler(
            f"{self.logdir}/{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.log")
        self.file_handler.setLevel(logging.DEBUG)
        self.logger.addHandler(self.file_handler)
        self.is_token_set = False
        self.api = API()
        self.init_api_token(token)
        formatter = logging.Formatter("[%(asctime)s - %(levelname)s] %(message)s")
        self.file_handler.setFormatter(formatter)
        atexit.register(self.cleanup)

    def log(self, msg: str, level: str = "debug") -> None:
        getattr(self.logger, level)(msg)

    def init_api_token(self, token: str or None) -> None:
        if token is None:
            return
        self.api.set_token(token)
        self.is_token_set = True

    def pushbullet_log(self, msg: str, algorithm: str = "MuZero") -> None:
        if not self.is_token_set:
            return
        try:
            self.api.send_note(f"{algorithm} training notification.", msg)
        except Exception as e:
            print(e)

    def clear_logdir(self):
        for file_name in os.listdir(self.logdir):
            os.remove(f"{self.logdir}/{file_name}")

    def cleanup(self) -> None:
        self.file_handler.close()
        self.logger.removeHandler(self.file_handler)


class LoggingMessageTemplates:

    @staticmethod
    def PITTING_START(name1: str, name2: str, num_games: int):
        return f"Starting pitting between {name1} and {name2} for {num_games} games."

    @staticmethod
    def PITTING_END(name1: str, name2: str, wins1: int, wins2: int, total: int, draws: int):
        return (f"Pitting ended between {name1} and {name2}. "
                f"Player 1 win frequency: {wins1 / total}. "
                f"Player 2 win frequency: {wins2 / total}. Draws: {draws}.")

    @staticmethod
    def SELF_PLAY_START(num_games: int):
        return f"Starting self play for {num_games} games."

    @staticmethod
    def SELF_PLAY_END(wins1: int, wins2: int, draws: int, not_zero_fn: callable):
        if wins1 is None or wins2 is None or draws is None:
            return "Self play ended. Results not available (This is expected if you are running MuZero)."
        return (f"Self play ended. Player 1 win frequency: {wins1 / (not_zero_fn(wins1 + wins2 + draws))}. "
                f"Player 2 win frequency: {wins2 / (not_zero_fn(wins1 + wins2 + draws))}. Draws: {draws}.")

    @staticmethod
    def NETWORK_TRAINING_START(num_epchs: int):
        return f"Starting network training for {num_epchs} epochs."

    @staticmethod
    def NETWORK_TRAINING_END(mean_loss: float):
        return f"Network training ended. Mean loss: {mean_loss}"

    @staticmethod
    def MODEL_REJECT(num_wins: float, update_threshold: float):
        return (f"!!! Model rejected, restoring previous version. Win rate: {num_wins}. "
                f"Update threshold: {update_threshold} !!!")

    @staticmethod
    def MODEL_ACCEPT(num_wins: float, update_threshold: float):
        return (
            f"!!! Model accepted, keeping current version. Win rate: {num_wins}. Update threshold: {update_threshold}"
            f" !!!")

    @staticmethod
    def TRAINING_START(num_iters: int):
        return f"Starting training for {num_iters} iterations."

    @staticmethod
    def TRAINING_END(args_used: dict):
        args_used_str = ""
        for key, value in args_used.items():
            args_used_str += f"{key}: {value}, "
        return f"Training ended. Args used: {args_used_str[:-2]}"

    @staticmethod
    def SAVED(type_: str, path: str):
        return f"Saved {type_} to {path}"

    @staticmethod
    def LOADED(type_: str, path: str):
        return f"Restored {type_} from {path}"

    @staticmethod
    def ITER_FINISHED_PSB(iter: int):
        return f"Iteration {iter} of the algorithm training finished!"

    @staticmethod
    def TRAINING_END_PSB():
        return "Algorithm Training finished, you can collect the results :)"



import math
import weakref
import numpy as np

from mu_alpha_zero.Game.tictactoe_game import TicTacToeGameManager as GameManager


class AlphaZeroNode:
    """
    This class defines a node in the search tree. It stores all the information needed for DeepMind's AlphaZero algorithm.
    """

    def __init__(self, current_player, select_probability=0, parent=None, times_visited_init=0):
        self.times_visited = times_visited_init
        self.was_init_with_zero_visits = times_visited_init == 0
        self.children = {}
        self.parent = weakref.ref(parent) if parent is not None else None
        self.select_probability = select_probability
        self.q = None
        self.current_player = current_player
        self.state = None
        self.total_value = 0

    def expand(self, state, action_probabilities) -> None:
        """
        Expands the newly visited node with the given action probabilities and state.
        :param state: np.ndarray of shape (board_size, board_size) representing the state current game board.
        :param action_probabilities: list of action probabilities for each action.
        :return: None
        """

        self.state = state.copy()  # copying here is probably not necessary.

        for action, probability in enumerate(action_probabilities):
            node = AlphaZeroNode(self.current_player * (-1), select_probability=probability, parent=self)
            self.children[action] = node

    def was_visited(self):
        return len(self.children) > 0

    def update_q(self, v):
        # Based on DeepMind's AlphaZero paper.
        if self.q is None:
            self.q = v
        else:
            self.q = (self.times_visited * self.q + v) / (self.times_visited + 1)

    def get_best_child(self, c=1.5):
        best_utc = -float("inf")
        best_child = None
        best_action = None
        valids_for_state = np.where(self.state != 0, -5, self.state)
        valids_for_state = np.where(valids_for_state == 0, 1, valids_for_state)
        utcs = []
        for action, child in self.children.items():
            action_ = np.unravel_index(action, self.state.shape)
            if valids_for_state[action_] != 1:
                continue
            child_utc = child.calculate_utc(c=c)
            utcs.append(child_utc)
            if child_utc > best_utc:
                best_utc = child_utc
                best_child = child
                best_action = action

        printable_children = [[child.state, child.times_visited, child.q, child.select_probability] for child in
                              self.children.values()]
        if best_child is None:
            # This was for testing purposes
            print("Best child is None. Possibly important info:\n", self.state, "\n",
                  valids_for_state, printable_children, self.was_visited(), utcs,
                  file=open("important_info"
                            ".txt", "w"))
            # raise Exception("Best child is None,terminating.")
        return best_child, best_action

    def calculate_utc(self, c=1.5):
        parent = self.parent()
        if self.q is None:
            # Inspiration taken from https://github.com/suragnair/alpha-zero-general/blob/master/MCTS.py
            utc = c * self.select_probability * math.sqrt(parent.times_visited + 1e-8)
        else:
            utc = self.q + c * (
                    self.select_probability * ((math.sqrt(parent.times_visited)) / (1 + self.times_visited)))

        return utc

    def get_self_value(self):
        # Not important for the algorithm, but might be useful for debugging.
        return self.total_value / self.times_visited if self.times_visited > 0 else 0

    def get_self_action_probabilities(self, tau=1.0, adjust=True):
        def adjust_probabilities(action_probabilities: dict, inner_tau=1.0) -> dict:
            """
            Selects a move from the action probabilities using either greedy or stochastic policy.
            The stochastic policy uses the tau parameter to adjust the probabilities. This is based on the
            temperature parameter in DeepMind's AlphaZero paper.
    
            :param action_probabilities: A dictionary containing the action probabilities in the form of {action_index: probability}.
            :param inner_tau: The temperature parameter. 0 for greedy, >0 for stochastic.
            :return: The selected move as an integer (index).
            """
            if inner_tau == 0:  # select greedy
                vals = [x for x in action_probabilities.values()]
                max_idx = vals.index(max(vals))
                probs = [0 for _ in range(len(vals))]
                probs[max_idx] = 1
                return dict(zip(action_probabilities.keys(), probs))
            # select stochastic
            moves, probabilities = zip(*action_probabilities.items())
            adjusted_probs = [prob ** (1 / inner_tau) for prob in probabilities]
            adjusted_probs_sum = sum(adjusted_probs)
            normalized_probs = [prob / adjusted_probs_sum for prob in adjusted_probs]
            return dict(zip(moves, normalized_probs))

        total_times_visited = self.times_visited
        action_probs = {}
        for action, child in self.children.items():
            action_probs[action] = child.times_visited / total_times_visited

        if adjust:
            return adjust_probabilities(action_probs, inner_tau=tau)
        else:
            return action_probs

    def get_latent(self):
        return self.state



import math

import numpy as np
import torch as th
import torch.nn.functional as F
from torch import nn
from torch.nn.functional import mse_loss

from mu_alpha_zero.AlphaZero.Network.nnet import AlphaZeroNet as PredictionNet
from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.network import GeneralMuZeroNetwork
from mu_alpha_zero.MuZero.utils import match_action_with_obs_batch, scale_reward_value
from mu_alpha_zero.config import MuZeroConfig


class MuZeroNet(th.nn.Module, GeneralMuZeroNetwork):
    def __init__(self, input_channels: int, dropout: float, action_size: int, num_channels: int, latent_size: int,
                 num_out_channels: int, linear_input_size: int):
        super(MuZeroNet, self).__init__()
        self.input_channels = input_channels
        self.dropout = dropout
        self.device = th.device("cuda" if th.cuda.is_available() else "cpu")
        self.action_size = action_size
        self.num_channels = num_channels
        self.latent_size = latent_size
        self.num_out_channels = num_out_channels
        self.linear_input_size = linear_input_size
        # self.action_embedding = th.nn.Embedding(action_size, 256)
        self.representation_network = RepresentationNet()
        self.dynamics_network = DynamicsNet(in_channels=257, num_channels=num_channels, dropout=dropout,
                                            latent_size=latent_size, out_channels=num_out_channels)
        # prediction outputs 6x6 latent state
        self.prediction_network = PredictionNet(in_channels=256, num_channels=num_channels, dropout=dropout,
                                                action_size=action_size, linear_input_size=linear_input_size)

    @classmethod
    def make_from_config(cls, config: MuZeroConfig):
        return cls(config.num_net_in_channels, config.net_dropout, config.net_action_size,
                   config.num_net_channels, config.net_latent_size, config.num_net_out_channels,
                   config.az_net_linear_input_size)

    def dynamics_forward(self, x: th.Tensor, predict: bool = False):
        if predict:
            return self.dynamics_network.predict(x)
        state, reward = self.dynamics_network(x)
        reward = scale_reward_value(reward)
        return state, reward

    def prediction_forward(self, x: th.Tensor, predict: bool = False):
        if predict:
            pi, v = self.prediction_network.predict(x, muzero=True)
            v = scale_reward_value(v[0][0])
            return pi, v
        pi, v = self.prediction_network(x, muzero=True)
        v = scale_reward_value(v)
        return pi, v

    def representation_forward(self, x: th.Tensor):
        x = self.representation_network(x)
        return x

    def make_fresh_instance(self):
        return MuZeroNet(self.input_channels, self.dropout, self.action_size, self.num_channels, self.latent_size,
                         self.num_out_channels, self.linear_input_size)

    def train_net(self, memory_buffer: GeneralMemoryBuffer, muzero_config: MuZeroConfig) -> tuple[float, list[float]]:
        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        losses = []
        K = muzero_config.K
        optimizer = th.optim.Adam(self.parameters(), lr=muzero_config.lr)
        for experience_batch, priorities in memory_buffer.batch_with_priorities(muzero_config.epochs,
                                                                                muzero_config.batch_size, K,
                                                                                alpha=muzero_config.alpha):
            if len(experience_batch) <= 1:
                continue
            pis, vs, rews_moves, states = zip(*experience_batch)
            rews = [x[0] for x in rews_moves]
            moves = [x[1] for x in rews_moves]
            states = th.tensor(np.array(states), dtype=th.float32, device=device) # .permute(0, 3, 1, 2) not used when using latent directly.
            pis = [list(x.values()) for x in pis]
            pis = th.tensor(np.array(pis), dtype=th.float32, device=device)
            vs = th.tensor(np.array(vs), dtype=th.float32, device=device).unsqueeze(1)
            rews = th.tensor(np.array(rews), dtype=th.float32, device=device).unsqueeze(1)
            # latent = self.representation_forward(states)
            latent = states  # Already predicted during search, this is to save space in RAM.
            pred_pis, pred_vs = self.prediction_forward(latent)
            latent = match_action_with_obs_batch(latent, moves)
            _, pred_rews = self.dynamics_forward(latent)
            priorities = priorities.to(device)
            balance_term = 1
            w = (1 / (len(priorities) * priorities)) ** muzero_config.beta
            w /= w.sum()
            loss_v = mse_loss(pred_vs, vs) * balance_term * w
            loss_pi = self.muzero_pi_loss(pred_pis, pis) * balance_term * w
            loss_r = mse_loss(pred_rews, rews) * balance_term * w
            loss = loss_v.sum() + loss_pi.sum() + loss_r.sum()
            losses.append(loss.item())
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        return sum(losses) / len(losses), losses

    def muzero_pi_loss(self, y_hat, y):
        return -th.sum(y * y_hat) / y.size()[0]

    def to_pickle(self, path: str):
        th.save(self, path)


class RepresentationNet(th.nn.Module):
    def __init__(self):
        super(RepresentationNet, self).__init__()
        self.device = th.device("cuda" if th.cuda.is_available() else "cpu")
        self.conv1 = th.nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=2, padding=1)
        self.residuals1 = th.nn.ModuleList([ResidualBlock(128) for _ in range(2)])
        self.conv2 = th.nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=2, padding=1)
        self.residuals2 = th.nn.ModuleList([ResidualBlock(256) for _ in range(3)])
        self.pool1 = th.nn.AvgPool2d(kernel_size=3, stride=2, padding=1)
        self.residuals3 = th.nn.ModuleList([ResidualBlock(256) for _ in range(3)])
        self.pool2 = th.nn.AvgPool2d(kernel_size=3, stride=2, padding=1)
        self.relu = th.nn.ReLU()

    def forward(self, x: th.Tensor):
        # x.unsqueeze(0)
        x = x.to(self.device)
        x = self.relu(self.conv1(x))
        for residual in self.residuals1:
            x = residual(x)
        x = self.relu(self.conv2(x))
        for residual in self.residuals2:
            x = residual(x)
        x = self.pool1(x)
        for residual in self.residuals3:
            x = residual(x)
        x = self.pool2(x)
        return x

    def trace(self):
        data = th.rand((1, 128, 8, 8))
        traced_script_module = th.jit.trace(self, data)
        return traced_script_module


class DynamicsNet(nn.Module):
    def __init__(self, in_channels, num_channels, dropout, latent_size, out_channels):
        super(DynamicsNet, self).__init__()
        self.out_channels = out_channels
        self.latent_size = int(math.sqrt(latent_size))

        # Convolutional layers
        self.conv1 = nn.Conv2d(in_channels, num_channels, 3, padding=1)
        self.bn1 = nn.BatchNorm2d(num_channels)
        self.conv2 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.bn2 = nn.BatchNorm2d(num_channels)
        self.conv3 = nn.Conv2d(num_channels, num_channels, 3, padding=1)
        self.bn3 = nn.BatchNorm2d(num_channels)
        self.conv4 = nn.Conv2d(num_channels, num_channels, 3)
        self.bn4 = nn.BatchNorm2d(num_channels)

        # Fully connected layers
        # 4608 (5x5) or 512 (3x3) or 32768 (10x10) or 18432 (8x8)
        self.fc1 = nn.Linear(8192, 1024)
        self.fc1_bn = nn.BatchNorm1d(1024)
        self.fc2 = nn.Linear(1024, 512)
        self.fc2_bn = nn.BatchNorm1d(512)

        self.dropout = nn.Dropout(dropout)

        # Output layers
        self.state_head = nn.Linear(512, latent_size * out_channels)  # state head
        self.reward_head = nn.Linear(512, 1)  # reward head

    def forward(self, x):
        # x = x.unsqueeze(0)
        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))
        x = F.relu(self.bn4(self.conv4(x)))

        x = x.view(x.size(0), -1)  # Flatten

        x = F.relu(self.fc1_bn(self.fc1(x)))
        x = self.dropout(x)

        x = F.relu(self.fc2_bn(self.fc2(x)))
        x = self.dropout(x)

        state = self.state_head(x)
        r = self.reward_head(x)

        return state, r

    def trace(self) -> th.jit.ScriptFunction:
        data = th.rand((1, 257, 6, 6)).to("cuda:0")
        traced_script_module = th.jit.trace(self, data)
        return traced_script_module

    @th.no_grad()
    def predict(self, x):
        state, r = self.forward(x)
        state = state.view(self.out_channels, self.latent_size, self.latent_size)
        r = scale_reward_value(r)
        return state, r.detach().cpu().numpy()


class ResidualBlock(th.nn.Module):
    def __init__(self, channels: int):
        super(ResidualBlock, self).__init__()
        self.convolution1 = th.nn.Conv2d(in_channels=channels, out_channels=channels, kernel_size=3, padding=1)
        self.bnorm1 = th.nn.BatchNorm2d(channels)
        self.convolution2 = th.nn.Conv2d(in_channels=channels, out_channels=channels, kernel_size=3, padding=1)
        self.bnorm2 = th.nn.BatchNorm2d(channels)
        self.relu = th.nn.ReLU()

    def forward(self, x):
        # x = x.unsqueeze(0)
        x_res = x
        convolved = self.convolution1(x)
        x = self.relu(self.bnorm1(convolved))
        x = self.bnorm2(self.convolution2(x))
        x += x_res
        x = self.relu(x)
        return x



from setuptools import setup
from setuptools.command.install import install
import subprocess


def get_requirements():
    with open("requirements.txt", "r") as file:
        return file.read().strip().split("\n")





setup(
    name="mu_alpha_zero_library",
    version="1.0.3",
    description="Library for running and training MuZero and AlphaZero models.",
    author="Skyr",
    install_requires=open("requirements.txt").read().strip().split("\n"),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    package_data={
        "mu_alpha_zero": ["*.txt", "*.root"]
    }

)



import os
import subprocess
import sys


def find_project_root() -> str:
    dir_ = os.path.dirname(os.path.abspath(__file__))
    os.chdir(dir_)
    while "root.root" not in os.listdir(dir_):
        if os.getcwd() == "/" or os.getcwd() == "C:\\":
            raise FileNotFoundError("Could not find project root.")
        os.chdir("..")
        dir_ = os.getcwd()
    return os.getcwd()


def not_zero(x):
    return x if x != 0 else 1


def get_python_home() -> str:
    return os.path.dirname(os.path.dirname(sys.executable))


def get_site_packages_path() -> str:
    output = subprocess.getoutput("pip show torch")
    return output.split("Location:")[1].split("Requires")[0].strip()


class OnlyUnique:
    def __init__(self, iterable: list = []):
        self.unique = []
        self.extend(iterable)

    def add(self, item):
        if item not in self.unique:
            self.unique.append(item)

    def extend(self, iterable):
        for item in iterable:
            self.add(item)

    def __len__(self):
        return len(self.unique)

    def get(self):
        return self.unique




def clear_disk_data():
    dir_ = f"{find_project_root()}/Pickles/Data"
    for file in os.listdir(dir_):
        os.remove(f"{dir_}/{file}")


def net_not_none(net):
    assert net is not None, (
        "Network is None, can't train/predict/pit. Make sure you initialize the network with either"
        "load_checkpoint or create_new method.")


def get_players() -> list[str]:
    path_prefix = find_project_root().replace("\\", "/").split("/")[-1]
    return [x for x in list(sys.modules[f"{path_prefix}.AlphaZero.Arena.players"].__dict__.keys()) if x.endswith("Player")]




import time
from typing import Type

from mu_alpha_zero.AlphaZero.Arena.players import Player
from mu_alpha_zero.Game.tictactoe_game import TicTacToeGameManager as GameManager
from mu_alpha_zero.General.arena import GeneralArena
from mu_alpha_zero.config import AlphaZeroConfig


class Arena(GeneralArena):
    def __init__(self, game_manager: GameManager, alpha_zero_config: AlphaZeroConfig, device):
        self.game_manager = game_manager
        self.device = device
        self.alpha_zero_config = alpha_zero_config

    def pit(self, player1: Type[Player], player2: Type[Player],
            num_games_to_play: int, num_mc_simulations: int, one_player: bool = False,
            start_player: int = 1, add_to_kwargs: dict or None = None, debug: bool = False) -> tuple[int, int, int]:
        """
        Pit two players against each other for a given number of games and gather the results.
        :param start_player: Which player should start the game.
        :param one_player: If True always only the first player will start the game.
        :param player1:
        :param player2:
        :param num_games_to_play:
        :param num_mc_simulations:
        :return: number of wins for player1, number of wins for player2, number of draws
        """
        # net_player_times = []
        # minimax_player_times = []
        results = {"wins_p1": 0, "wins_p2": 0, "draws": 0}
        tau = self.alpha_zero_config.arena_tau
        if one_player:
            num_games_per_player = num_games_to_play
        else:
            num_games_per_player = num_games_to_play // 2

        for game in range(num_games_to_play):
            if game < num_games_per_player:
                current_player = start_player
            else:
                current_player = -start_player

            kwargs = {"num_simulations": num_mc_simulations, "current_player": current_player, "device": self.device,
                      "tau": tau}
            if add_to_kwargs is not None:
                kwargs.update(add_to_kwargs)
            state = self.game_manager.reset()
            if player1.name == "NetworkPlayer":
                player1.monte_carlo_tree_search.step_root(None)
            if player2.name == "NetworkPlayer":
                player2.monte_carlo_tree_search.step_root(None)
            # time.sleep(0.01)
            while True:
                self.game_manager.render()
                if current_player == 1:
                    # p1_time = time.time()
                    move = player1.choose_move(state, **kwargs)
                    # net_player_times.append(time.time() - p1_time)
                else:
                    # p2_time = time.time()
                    move = player2.choose_move(state, **kwargs)
                    # minimax_player_times.append(time.time() - p2_time)
                self.game_manager.play(current_player, move)
                state = self.game_manager.get_board()
                status = self.game_manager.game_result(current_player, state)
                self.game_manager.render()

                if status is not None:
                    # time.sleep(3)
                    # print(state)
                    if status == 1:
                        if current_player == 1:
                            results["wins_p1"] += 1
                        else:
                            results["wins_p2"] += 1

                    elif status == -1:
                        if current_player == 1:
                            results["wins_p2"] += 1
                        else:
                            results["wins_p1"] += 1
                    else:
                        results["draws"] += 1

                    # if debug:
                    #     self.wait_keypress()
                    if (player1.name == "HumanPlayer" or player2.name == "HumanPlayer") and debug:
                        time.sleep(0.2)
                    break

                current_player *= -1
                kwargs["current_player"] = current_player
        # print(f"Average net player time: {sum(net_player_times) / len(net_player_times)}")
        # print(f"Average minimax player time: {sum(minimax_player_times) / len(minimax_player_times)}")
        # print(f"Net player times: {net_player_times}")
        # print(f"Minimax player times: {minimax_player_times}")
        return results["wins_p1"], results["wins_p2"], results["draws"]

    def wait_keypress(self):
        inpt = input("Press any key to continue...")
        return inpt



from abc import abstractmethod, ABC
from typing import Type

from mu_alpha_zero.AlphaZero.Arena.players import Player


class GeneralArena(ABC):
    @abstractmethod
    def pit(self, player1: Type[Player], player2: Type[Player],
            num_games_to_play: int, num_mc_simulations: int, one_player: bool = False,
            start_player: int = 1):
        pass
